@@ -20,33 +20,35 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+#include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <signal.h>
-#include <unistd.h>
+#include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
 #include <errno.h>
 #include <getopt.h>
-#include <time.h>
-#include <sys/poll.h>
-#include <grp.h>
-#include <syslog.h>
+#include <stdarg.h>
 
 #include "acpid.h"
-#include "ud_socket.h"
+#include "event.h"
+#include "connection_list.h"
+#include "proc.h"
+#include "sock.h"
+#include "input_layer.h"
+#include "inotify_handler.h"
+#include "netlink.h"
 
 static int handle_cmdline(int *argc, char ***argv);
 static void close_fds(void);
 static int daemonize(void);
-static int open_log(void);
+static void open_log(void);
+static int std2null(void);
 static int create_pidfile(void);
 static void clean_exit(int sig);
-static void clean_exit_with_status(int status);
 static void reload_conf(int sig);
-static char *read_line(int fd);
 
 /* global debug level */
 int acpid_debug;
@@ -54,27 +56,17 @@ int acpid_debug;
 /* do we log event info? */
 int logevents;
 
-/* the number of non-root clients that are connected */
-int non_root_clients;
-
-static const char *progname;
+const char *progname;
 static const char *confdir = ACPID_CONFDIR;
-static const char *eventfile = ACPID_EVENTFILE;
-static const char *socketfile = ACPID_SOCKETFILE;
 static const char *lockfile = ACPID_LOCKFILE;
 static int nosocket;
-static const char *socketgroup;
-static mode_t socketmode = ACPID_SOCKETMODE;
 static int foreground;
 static const char *pidfile = ACPID_PIDFILE;
-static int clientmax = ACPID_CLIENTMAX;
+static int netlink;
 
 int
 main(int argc, char **argv)
 {
-	int event_fd;
-	int sock_fd = -1; /* init to avoid a compiler warning */
-
 	/* learn who we really are */
 	progname = (const char *)strrchr(argv[0], '/');
 	progname = progname ? (progname + 1) : argv[0];
@@ -85,78 +77,32 @@ main(int argc, char **argv)
 	/* close any extra file descriptors */
 	close_fds();
 
-	/* actually open the event file */
-	event_fd = open(eventfile, O_RDONLY);
-	if (event_fd < 0) {
-		fprintf(stderr, "%s: can't open %s: %s\n", progname, 
-			eventfile, strerror(errno));
-		exit(EXIT_FAILURE);
-	}
-	fcntl(event_fd, F_SETFD, FD_CLOEXEC);
-
-/*
- * if there is data, and the kernel is NOT broken, this eats 1 byte.  We
- * can't have that.  This is ifdef'ed out on the assumption that old kernels
- * are out of popular use, by now.
- */
-#ifdef TEST_FOR_BAD_KERNELS
-	/*
-	 * Older kernels did not support read() properly or poll() at all
-	 * Check that the kernel supports the proper semantics, or die.
-	 *
-	 * Good kernels will respect O_NONBLOCK and return -1.  Bad kernels
-	 * will ignore O_NONBLOCK and return 0.  Really bad kernels will block
-	 * and overflow the buffer.  Can't deal with the really bad ones.
-	 */
+	/* open the log */
+	open_log();
+	
+	if (!netlink)
 	{
-		int fl;
-		char buf;
+		/* open the acpi event file in the proc fs */
+		/* if the open fails, try netlink */
+		if (open_proc())
+			netlink = 1;
+	}
 
-		fl = fcntl(event_fd, F_GETFL);
-		fcntl(event_fd, F_SETFL, fl | O_NONBLOCK);
-		if (read(event_fd, &buf, 1) == 0) {
-			fprintf(stderr, 
-				"%s: this kernel does not support proper "
-				"event file handling.\n"
-				"Please get the patch from "
-				"http://acpid.sourceforge.net\n", 
-				progname);
-			exit(EXIT_FAILURE);
-		}
-		fcntl(event_fd, F_SETFL, fl);
+	if (netlink)
+	{
+		/* open the input layer */
+		open_input();
+
+		/* watch for new input layer devices */
+		open_inotify();
+
+		/* open netlink */
+		open_netlink();
 	}
-#endif
 
 	/* open our socket */
 	if (!nosocket) {
-		sock_fd = ud_create_socket(socketfile);
-		if (sock_fd < 0) {
-			fprintf(stderr, "%s: can't open socket %s: %s\n",
-				progname, socketfile, strerror(errno));
-			exit(EXIT_FAILURE);
-		}
-		fcntl(sock_fd, F_SETFD, FD_CLOEXEC);
-		chmod(socketfile, socketmode);
-		if (socketgroup) {
-			struct group *gr;
-			struct stat buf;
-			gr = getgrnam(socketgroup);
-			if (!gr) {
-				fprintf(stderr, "%s: group %s does not exist\n",
-					progname, socketgroup);
-				exit(EXIT_FAILURE);
-			}
-			if (stat(socketfile, &buf) < 0) {
-				fprintf(stderr, "%s: can't stat %s\n",
-					progname, socketfile);
-				exit(EXIT_FAILURE);
-			}
-			if (chown(socketfile, buf.st_uid, gr->gr_gid) < 0) {
-				fprintf(stderr, "%s: chown(): %s\n",
-					progname, strerror(errno));
-				exit(EXIT_FAILURE);
-			}
-		}
+		open_sock();
 	}
 
 	/* if we're running in foreground, we don't daemonize */
@@ -165,11 +111,13 @@ main(int argc, char **argv)
 			exit(EXIT_FAILURE);
 	}
 
-	/* open the log */
-	if (open_log() < 0) {
+	/* redirect standard files to /dev/null */
+	if (std2null() < 0) {
 		exit(EXIT_FAILURE);
 	}
-	acpid_log(LOG_INFO, "starting up\n");
+	
+	acpid_log(LOG_INFO, "starting up with %s\n",
+		netlink ? "netlink and the input layer" : "proc fs");
 
 	/* trap key signals */
 	signal(SIGHUP, reload_conf);
@@ -188,128 +136,53 @@ main(int argc, char **argv)
 		exit(EXIT_FAILURE);
 	}
 
-	/* main loop */
 	acpid_log(LOG_INFO, "waiting for events: event logging is %s\n",
 	    logevents ? "on" : "off");
-	while (1) {
-		struct pollfd ar[2];
-		int r;
-		int fds = 0;
 
-		/* poll for the socket and the event file */
-		ar[0].fd = event_fd; ar[0].events = POLLIN; fds++;
-		if (!nosocket) {
-			ar[1].fd = sock_fd; ar[1].events = POLLIN; fds++;
-		}
-		r = poll(ar, fds, -1);
+	/* main loop */
+	while (1)
+	{
+		fd_set readfds;
+		int nready;
+		int i;
+		struct connection *p;
 
-		if (r < 0 && errno == EINTR) {
+		/* it's going to get clobbered, so use a copy */
+		readfds = *get_fdset();
+
+		/* wait on data */
+		nready = select(get_highestfd() + 1, &readfds, NULL, NULL, NULL);
+
+		if (nready < 0  &&  errno == EINTR) {
 			continue;
-		} else if (r < 0) {
-			acpid_log(LOG_ERR, "poll(): %s\n", strerror(errno));
+		} else if (nready < 0) {
+			acpid_log(LOG_ERR, "select(): %s\n", strerror(errno));
 			continue;
 		}
 
 		/* house keeping */
 		acpid_close_dead_clients();
 
-		/* was it an event? */
-		if (ar[0].revents) {
-			char *event;
-			struct stat trash;
-			int fexists;
+		/* for each connection */
+		for (i = 0; i <= get_number_of_connections(); ++i)
+		{
+			int fd;
 
-			/* check for existence of a lockfile */
-			fexists = (stat(lockfile, &trash) == 0);
+			p = get_connection(i);
 
-			/* this shouldn't happen */
-			if (!ar[0].revents & POLLIN) {
-				acpid_log(LOG_DEBUG,
-				    "odd, poll set flags 0x%x\n",
-				    ar[0].revents);
-				continue;
-			}
-
-			/* read an event */
-			event = read_line(event_fd);
-
-			/* if we're locked, don't process the event */
-			if (fexists) {
-				if (logevents) {
-					acpid_log(LOG_INFO,
-					    "lockfile present, not processing "
-					    "event \"%s\"\n", event);
-				}
-				continue;
-			}
-
-			/* handle the event */
-			if (event) {
-				if (logevents) {
-					acpid_log(LOG_INFO,
-					    "received event \"%s\"\n", event);
-				}
-				acpid_handle_event(event);
-				if (logevents) {
-					acpid_log(LOG_INFO,
-					    "completed event \"%s\"\n", event);
-				}
-			} else if (errno == EPIPE) {
-				acpid_log(LOG_WARNING,
-				    "events file connection closed\n");
+			/* if this connection is invalid, bail */
+			if (!p)
 				break;
-			} else {
-				static int nerrs;
-				if (++nerrs >= ACPID_MAX_ERRS) {
-					acpid_log(LOG_ERR,
-					    "too many errors reading "
-					    "events file - aborting\n");
-					break;
-				}
-			}
-		}
 
-		/* was it a new connection? */
-		if (!nosocket && ar[1].revents) {
-			int cli_fd;
-			struct ucred creds;
-			char buf[32];
-			static int accept_errors;
+			/* get the file descriptor */
+			fd = p->fd;
 
-			/* this shouldn't happen */
-			if (!ar[1].revents & POLLIN) {
-				acpid_log(LOG_DEBUG,
-				    "odd, poll set flags 0x%x\n",
-				    ar[1].revents);
-				continue;
+			/* if this file descriptor has data waiting */
+			if (FD_ISSET(fd, &readfds))
+			{
+				/* delegate to this connection's process function */
+				p->process(fd);
 			}
-
-			/* accept and add to our lists */
-			cli_fd = ud_accept(sock_fd, &creds);
-			if (cli_fd < 0) {
-				acpid_log(LOG_ERR, "can't accept client: %s\n",
-				    strerror(errno));
-				accept_errors++;
-				if (accept_errors >= 5) {
-					acpid_log(LOG_ERR, "giving up\n");
-					clean_exit_with_status(EXIT_FAILURE);
-				}
-				continue;
-			}
-			accept_errors = 0;
-			if (creds.uid != 0 && non_root_clients >= clientmax) {
-				close(cli_fd);
-				acpid_log(LOG_ERR,
-				    "too many non-root clients\n");
-				continue;
-			}
-			if (creds.uid != 0) {
-				non_root_clients++;
-			}
-			fcntl(cli_fd, F_SETFD, FD_CLOEXEC);
-			snprintf(buf, sizeof(buf)-1, "%d[%d:%d]",
-				creds.pid, creds.uid, creds.gid);
-			acpid_add_client(cli_fd, buf);
 		}
 	}
 
@@ -337,6 +210,7 @@ handle_cmdline(int *argc, char ***argv)
 		{"nosocket", 1, 0, 'S'},
 		{"pidfile", 1, 0, 'p'},
 		{"lockfile", 1, 0, 'L'},
+		{"netlink", 0, 0, 'n'},
 		{"version", 0, 0, 'v'},
 		{"help", 0, 0, 'h'},
 		{NULL, 0, 0, 0},
@@ -353,7 +227,8 @@ handle_cmdline(int *argc, char ***argv)
 		"Use the specified socket file.",	/* socketfile */
 		"Do not listen on a UNIX socket (overrides -s).",/* nosocket */
 		"Use the specified PID file.",		/* pidfile */
-		"Use the specified lockfile to stop processing.", /* pidfile */
+		"Use the specified lockfile to stop processing.", /* lockfile */
+		"Force netlink/input layer mode. (overrides -e)", /* netlink */
 		"Print version information.",		/* version */
 		"Print this message.",			/* help */
 	};
@@ -364,7 +239,7 @@ handle_cmdline(int *argc, char ***argv)
 	for (;;) {
 		int i;
 		i = getopt_long(*argc, *argv,
-		    "c:C:de:flg:m:s:Sp:L:vh", opts, NULL);
+		    "c:C:de:flg:m:s:Sp:L:nvh", opts, NULL);
 		if (i == -1) {
 			break;
 		}
@@ -406,6 +281,9 @@ handle_cmdline(int *argc, char ***argv)
 		case 'L':
 			lockfile = optarg;
 			break;
+		case 'n':
+			netlink = 1;
+			break;
 		case 'v':
 			printf(PACKAGE "-" VERSION "\n");
 			exit(EXIT_SUCCESS);
@@ -452,63 +330,82 @@ close_fds(void)
 static int
 daemonize(void)
 {
-	switch(fork()) {
-	case -1:
-		fprintf(stderr, "%s: fork: %s\n", progname, strerror(errno));
+	pid_t pid, sid;
+
+    /* already a daemon */
+    if ( getppid() == 1 ) return 0;
+
+	/* fork off the parent process */
+	pid = fork();
+	if (pid < 0) {
+		acpid_log(LOG_ERR, "fork: %s\n", strerror(errno));
 		return -1;
-	case 0:
-		/* child */
-		break;
-	default:
-		/* parent */
+	}
+	/* if we got a good PID, then we can exit the parent process */
+	if (pid > 0) {
 		exit(EXIT_SUCCESS);
 	}
 
-	/* disconnect */
-	setsid();
+	/* at this point we are executing as the child process */
+
+	/* change the umask to something predictable instead of inheriting */
+	/* whatever from the parent */
 	umask(0);
 
-	/* get outta the way */
+	/* create a new SID for the child process and */
+	/* detach the process from the parent (normally a shell) */
+	sid = setsid();
+	if (sid < 0) {
+		acpid_log(LOG_ERR, "setsid: %s\n", strerror(errno));
+		return -1;
+	}
+
+    /* Change the current working directory.  This prevents the current
+       directory from being locked; hence not being able to remove it. */
 	if (chdir("/") < 0) {
-		fprintf(stderr, "%s: chdir(\"/\"): %s\n", progname,
-		        strerror(errno));
+		acpid_log(LOG_ERR, "chdir(\"/\"): %s\n", strerror(errno));
 		return -1;
 	}
 
 	return 0;
 }
 
-static int
+static void
 open_log(void)
 {
-	int nullfd;
 	int log_opts;
 
-	/* open /dev/null */
-	nullfd = open("/dev/null", O_RDWR);
-	if (nullfd < 0) {
-		fprintf(stderr, "%s: can't open %s: %s\n", progname, 
-			"/dev/null", strerror(errno));
-		return -1;
-	}
-
+	/* open the syslog */
 	log_opts = LOG_CONS|LOG_NDELAY;
 	if (acpid_debug) {
 		log_opts |= LOG_PERROR;
 	}
 	openlog(PACKAGE, log_opts, LOG_DAEMON);
+}
+
+static int
+std2null(void)
+{
+	int nullfd;
+
+	/* open /dev/null */
+	nullfd = open("/dev/null", O_RDWR);
+	if (nullfd < 0) {
+		acpid_log(LOG_ERR, "can't open /dev/null: %s\n", strerror(errno));
+		return -1;
+	}
 
 	/* set up stdin, stdout, stderr to /dev/null */
 	if (dup2(nullfd, STDIN_FILENO) != STDIN_FILENO) {
-		fprintf(stderr, "%s: dup2: %s\n", progname, strerror(errno));
+		acpid_log(LOG_ERR, "dup2() stdin: %s\n", strerror(errno));
 		return -1;
 	}
 	if (!acpid_debug && dup2(nullfd, STDOUT_FILENO) != STDOUT_FILENO) {
-		fprintf(stderr, "%s: dup2: %s\n", progname, strerror(errno));
+		acpid_log(LOG_ERR, "dup2() stdout: %s\n", strerror(errno));
 		return -1;
 	}
 	if (!acpid_debug && dup2(nullfd, STDERR_FILENO) != STDERR_FILENO) {
-		fprintf(stderr, "%s: dup2: %s\n", progname, strerror(errno));
+		acpid_log(LOG_ERR, "dup2() stderr: %s\n", strerror(errno));
 		return -1;
 	}
 
@@ -547,7 +444,7 @@ create_pidfile(void)
 	return -1;
 }
 
-static void
+void
 clean_exit_with_status(int status)
 {
 	acpid_cleanup_rules(1);
@@ -578,61 +475,29 @@ acpid_log(int level, const char *fmt, ...)
 {
 	va_list args;
 
-	va_start(args, fmt);
-	vsyslog(level, fmt, args);
-	va_end(args);
+	if (level == LOG_DEBUG) {
+		/* if "-d" has been specified */
+		if (acpid_debug) {
+			/* send debug output to stderr */
+			va_start(args, fmt);
+			vfprintf(stderr, fmt, args);
+			va_end(args);
+		}
+	} else {
+		va_start(args, fmt);
+		vsyslog(level, fmt, args);
+		va_end(args);
+	}
 
 	return 0;
 }
 
-/* 
- * This depends on fixes in linux ACPI after 2.4.8
- */
-#define MAX_BUFLEN	1024
-static char *
-read_line(int fd)
+int
+locked()
 {
-	static char *buf;
-	int buflen = 64;
-	int i = 0;
-	int r;
-	int searching = 1;
+	struct stat trash;
 
-	while (searching) {
-		buf = realloc(buf, buflen);
-		if (!buf) {
-			acpid_log(LOG_ERR, "malloc(%d): %s\n",
-				buflen, strerror(errno));
-			return NULL;
-		}
-		memset(buf+i, 0, buflen-i);
-
-		while (i < buflen) {
-			r = read(fd, buf+i, 1);
-			if (r < 0 && errno != EINTR) {
-				/* we should do something with the data */
-				acpid_log(LOG_ERR, "read(): %s\n",
-					strerror(errno));
-				return NULL;
-			} else if (r == 0) {
-				/* signal this in an almost standard way */
-				errno = EPIPE;
-				return NULL;
-			} else if (r == 1) {
-				/* scan for a newline */
-				if (buf[i] == '\n') {
-					searching = 0;
-					buf[i] = '\0';
-					break;
-				}
-				i++;
-			}
-		}
-		if (buflen >= MAX_BUFLEN) {
-			break;
-		} 
-		buflen *= 2;
-	}
-
-	return buf;
+	/* check for existence of a lockfile */
+	return (stat(lockfile, &trash) == 0);
 }
+
