@@ -1,5 +1,5 @@
 /*
- *  event.c - ACPI daemon
+ *  event.c - ACPI daemon event handler
  *
  *  Copyright (C) 2000 Andrew Henroid
  *  Copyright (C) 2001 Sun Microsystems (thockin@sun.com)
@@ -36,6 +36,7 @@
 #include <signal.h>
 
 #include "acpid.h"
+#include "sock.h"
 #include "ud_socket.h"
 
 /*
@@ -92,6 +93,8 @@ acpid_read_conf(const char *confdir)
 	struct dirent *dirent;
 	char *file = NULL;
 	int nrules = 0;
+	regex_t preg;
+	int rc = 0;
 
 	lock_rules();
 
@@ -103,21 +106,31 @@ acpid_read_conf(const char *confdir)
 		return -1;
 	}
 
+	/* Compile the regular expression.  This is based on run-parts(8). */
+	rc = regcomp(&preg, "^[a-zA-Z0-9_-]+$", RULE_REGEX_FLAGS);
+	if (rc) {
+		acpid_log(LOG_ERR, "regcomp(): %d\n", rc);
+		unlock_rules();
+		return -1;
+	}
+
 	/* scan all the files */
 	while ((dirent = readdir(dir))) {
 		int len;
 		struct rule *r;
 		struct stat stat_buf;
 
-		if (dirent->d_name[0] == '.')
-			continue; /* skip dotfiles */
-
 		len = strlen(dirent->d_name);
 
-		if (dirent->d_name[len - 1] == '~') {
-			continue; /* skip editor backup files */
+		/* skip any files that don't match the run-parts convention */
+		if (regexec(&preg, dirent->d_name, 0, NULL, 0) != 0) {
+			acpid_log(LOG_INFO, "skipping conf file %s/%s\n", 
+				confdir, dirent->d_name);
+			continue;
 		}
 
+		/* Compute the length of the full path name adding one for */
+		/* the slash and one more for the NULL. */
 		len += strlen(confdir) + 2;
 
 		file = malloc(len);
@@ -136,7 +149,7 @@ acpid_read_conf(const char *confdir)
 			continue; /* keep trying the rest of the files */
 		}
 		if (!S_ISREG(stat_buf.st_mode)) {
-			acpid_log(LOG_DEBUG, "skipping non-file %s\n", file);
+			acpid_log(LOG_INFO, "skipping non-file %s\n", file);
 			free(file);
 			continue; /* skip non-regular files */
 		}
@@ -206,9 +219,7 @@ parse_file(const char *file)
 	int line = 0;
 	struct rule *r;
 
-	if (acpid_debug) {
-		acpid_log(LOG_DEBUG, "parsing conf file %s\n", file);
-	}
+	acpid_log(LOG_DEBUG, "parsing conf file %s\n", file);
 
 	fp = fopen(file, "r");
 	if (!fp) {
@@ -310,10 +321,7 @@ parse_file(const char *file)
 		}
 	}
 	if (!r->event || !r->action.cmd) {
-		if (acpid_debug) {
-			acpid_log(LOG_DEBUG, "skipping incomplete file %s\n",
-			    file);
-		}
+		acpid_log(LOG_INFO, "skipping incomplete file %s\n", file);
 		free_rule(r);
 		fclose(fp);
 		return NULL;
@@ -523,8 +531,8 @@ acpid_handle_event(const char *event)
 			struct rule *pnext = p->next;
 			if (!regexec(p->event, event, 0, NULL, 0)) {
 				/* a match! */
-				if (acpid_debug && logevents) {
-					acpid_log(LOG_DEBUG,
+				if (logevents) {
+					acpid_log(LOG_INFO,
 					    "rule from %s matched\n",
 					    p->origin);
 				}
@@ -540,7 +548,7 @@ acpid_handle_event(const char *event)
 				}
 			} else {
 				if (acpid_debug >= 3 && logevents) {
-					acpid_log(LOG_DEBUG,
+					acpid_log(LOG_INFO,
 					    "rule from %s did not match\n",
 					    p->origin);
 				}
@@ -551,8 +559,8 @@ acpid_handle_event(const char *event)
 
 	unlock_rules();
 
-	if (acpid_debug && logevents) {
-		acpid_log(LOG_DEBUG, "%d total rule%s matched\n",
+	if (logevents) {
+		acpid_log(LOG_INFO, "%d total rule%s matched\n",
 			nrules, (nrules == 1)?"":"s");
 	}
 
